@@ -6,9 +6,11 @@
 """
 from typing import Iterable, Set
 from pathlib import Path
+from collections import namedtuple
 from urllib.parse import quote_plus
-from .catalog import register, rt_register
-from .models_base import ApiItem, IndexApiItem, ConfigItem, IndexConfigItem, RealtimeItem, ApiPath, IdName
+from .catalog import register, op_register
+from .models_base import (ApiItem, IndexApiItem, ConfigItem, IndexConfigItem, RealtimeItem, BulkStatsItem,
+                          BulkStateItem, ApiPath, IdName)
 
 
 #
@@ -187,7 +189,7 @@ class ControlInventory(IndexApiItem):
 #
 class DeviceConfig(ConfigItem):
     api_path = ApiPath('template/config/attached', None, None, None)
-    store_path = ('device_configs', )
+    store_path = ('device_configs',)
     store_file = '{item_name}.txt'
 
     def save(self, node_dir, ext_name=False, item_name=None, item_id=None):
@@ -242,7 +244,7 @@ class CliOrFeatureApiPath:
         return self.api_path_cli if is_cli_template else self.api_path_feature
 
 
-# Set of device types that use cedge template class. Updated as of vManage 20.3.2
+# Set of device types that use cedge template class. Updated as of vManage 20.6.1
 CEDGE_SET = {
     "vedge-CSR-1000v", "vedge-ISR-4331", "vedge-ISR-4431", "vedge-ISR-4461", "vedge-ISR-4451-X",
     "vedge-C8300-1N1S-4T2X", "vedge-IR-1101", "vedge-C8300-1N1S-6T", "vedge-ISRv", "vedge-ISR-4321", "vedge-ISR-4351",
@@ -257,8 +259,19 @@ CEDGE_SET = {
     "vedge-C1127-8PLTEP", "vedge-C1161-8P", "vedge-C1117-4P", "vedge-C1117-4PM", "vedge-C1117-4PLTEEA",
     "vedge-C1126X-8PLTEP", "vedge-C1127X-8PLTEP", "vedge-C1121X-8PLTEPW", "vedge-C1127X-8PMLTEP", "vedge-C1127-8PMLTEP",
     "vedge-C1117-4PLTELA", "vedge-nfvis-ENCS5400", 'vedge-C1113-8PLTEW', 'vedge-ESR-6300', "vedge-C8300-2N2S-6T",
-    "vedge-C8300-2N2S-4T2X", "vedge-C1117-4PMLTEEA"
+    "vedge-C8300-2N2S-4T2X", "vedge-C1117-4PMLTEEA", "vedge-C1113-8PW", "vedge-ISR1100-4GLTENA-XE",
+    "vedge-C1117-4PMLTEEAWE", "vedge-ASR-1006-X", "vedge-ISR1100X-6G-XE", "vedge-C1113-8PM", "vedge-C1116-4PWE",
+    "vedge-IR-1835", "vedge-C8500L-8S4X", "vedge-C1113-8PLTEEAW", "vedge-C1117-4PLTELAWZ", "vedge-C1112-8PWE",
+    "cellular-gateway-CG522-E", "vedge-C1117-4PLTEEAW", "vedge-C1116-4PLTEEAWE", "vedge-C1112-8PLTEEAWE",
+    "vedge-C1113-8PLTELAWZ", "cellular-gateway-CG418-E", "vedge-ISR1100-6G-XE", "vedge-C1113-8PMWE", "vedge-C1111-4PW",
+    "vedge-C1113-8PLTELA", "vedge-C1118-8P", "vedge-C1112-8P", "vedge-ISR1100-4G-XE", "vedge-IR-1833",
+    "vedge-ISR1100X-4G-XE", "vedge-C1117-4PMWE", "vedge-IR-1821", "vedge-C1161-8PLTEP", "vedge-ISR1100-4GLTEGB-XE",
+    "vedge-nfvis-C8200-UCPE", "vedge-C8000V", "vedge-C1117-4PW", "vedge-C8200-1N-4T", "vedge-C1112-8PLTEEA",
+    "vedge-C1113-8P", "vedge-IR-1831", "vedge-C8200L-1N-4T", "vedge-nfvis-C8200-UCPEVM", "vedge-C8200L-1N-4T",
+    "vedge-nfvis-C8200-UCPEVM", "vedge-IR-8340"
 }
+# Software devices. Updated as of vManage 20.6.1
+SOFT_EDGE_SET = {"vedge-CSR-1000v", "vedge-C8000V", "vedge-cloud", "vmanage", "vsmart"}
 
 
 class DeviceTemplate(ConfigItem):
@@ -269,9 +282,12 @@ class DeviceTemplate(ConfigItem):
     store_path = ('device_templates', 'template')
     store_file = '{item_name}.json'
     name_tag = 'templateName'
-    post_filtered_tags = ('feature', )
+    post_filtered_tags = ('feature',)
+    # templateClass, deviceRole, draftMode, templateId and copyEdited are new tags in 20.x+, adding to skip diff to not
+    # trigger updates when restore --update is done between pre 20.x workdir and post 20.x vManage.
     skip_cmp_tag_set = {'createdOn', 'createdBy', 'lastUpdatedBy', 'lastUpdatedOn', '@rid', 'owner', 'infoTag',
-                        'templateAttached', 'templateConfigurationEdited'}
+                        'templateAttached', 'templateConfigurationEdited', 'templateClass', 'deviceRole', 'draftMode',
+                        'templateId', 'copyEdited'}
 
     @property
     def is_type_cli(self) -> bool:
@@ -371,13 +387,15 @@ class DeviceTemplateValues(ConfigItem):
 
 class FeatureTemplate(ConfigItem):
     api_path = ApiPath('template/feature/object', 'template/feature')
-    store_path = ('feature_templates', )
+    store_path = ('feature_templates',)
     store_file = '{item_name}.json'
     id_tag = 'templateId'
     name_tag = 'templateName'
     type_tag = 'templateType'
+    # gTemplateClass is new in 20.x, adding skip diff to not trigger updates when restore --update is done between
+    # pre 20.x workdir and post 20.x vManage.
     skip_cmp_tag_set = {'createdOn', 'createdBy', 'lastUpdatedBy', 'lastUpdatedOn', '@rid', 'owner', 'infoTag',
-                        'devicesAttached', 'attachedMastersCount'}
+                        'devicesAttached', 'attachedMastersCount', 'gTemplateClass'}
 
     @property
     def device_types(self) -> Set[str]:
@@ -491,6 +509,9 @@ class PolicySecurity(ConfigItem):
     store_file = '{item_name}.json'
     name_tag = 'policyName'
     type_tag = 'policyType'
+    # policyUseCase, policyMode are new tags in 20.x+, adding to skip diff to not trigger updates when restore --update
+    # is done between pre 20.x workdir and post 20.x vManage.
+    skip_cmp_tag_set = {'policyUseCase', 'policyMode'}
 
 
 @register('policy_security', 'security policy', PolicySecurity)
@@ -558,8 +579,10 @@ class PolicyDef(ConfigItem):
     id_tag = 'definitionId'
     name_tag = 'name'
     type_tag = 'type'
+    # mode is new tag in 20.x+, adding to skip diff to not trigger updates when restore --update is done between pre
+    # 20.x workdir and post 20.x vManage.
     skip_cmp_tag_set = {'lastUpdated', 'referenceCount', 'references', 'activatedId', 'isActivatedByVsmart',
-                        'owner', 'infoTag'}
+                        'owner', 'infoTag', 'mode'}
 
 
 # Policy definition index base class
@@ -864,6 +887,17 @@ class PolicyDefPriisdnPortIndex(PolicyDefIndex):
     store_file = 'policy_definitions_priisdnport.json'
 
 
+class PolicyDefRuleSet(PolicyDef):
+    api_path = ApiPath('template/policy/definition/ruleset')
+    store_path = ('policy_definitions', 'RuleSet')
+
+
+@register('policy_definition', 'rule-set security policy definition', PolicyDefRuleSet, min_version='20.4')
+class PolicyDefRuleSetIndex(PolicyDefIndex):
+    api_path = ApiPath('template/policy/definition/ruleset', None, None, None)
+    store_file = 'policy_definitions_ruleset.json'
+
+
 #
 # Policy lists
 #
@@ -960,17 +994,6 @@ class PolicyListUmbrellaDataIndex(PolicyListIndex):
     store_file = 'policy_lists_umbrelladata.json'
 
 
-class PolicyListPrefix(PolicyList):
-    api_path = ApiPath('template/policy/list/prefix')
-    store_path = ('policy_lists', 'Prefix')
-
-
-@register('policy_list', 'prefix list', PolicyListPrefix)
-class PolicyListPrefixIndex(PolicyListIndex):
-    api_path = ApiPath('template/policy/list/prefix', None, None, None)
-    store_file = 'policy_lists_prefix.json'
-
-
 class PolicyListSite(PolicyList):
     api_path = ApiPath('template/policy/list/site')
     store_path = ('policy_lists', 'Site')
@@ -995,6 +1018,7 @@ class PolicyListExtcommunityIndex(PolicyListIndex):
 
 # Data Prefix All (template/policy/list/dataprefixall) was purposely not included as it seems to collide with, meaning
 # error, Data Prefix (template/policy/list/dataprefix).
+# Data Prefix FQDN (template/policy/list/dataprefixfqdn) was also not included for the same reason.
 class PolicyListDataprefix(PolicyList):
     api_path = ApiPath('template/policy/list/dataprefix')
     store_path = ('policy_lists', 'DataPrefix')
@@ -1006,6 +1030,17 @@ class PolicyListDataprefixIndex(PolicyListIndex):
     store_file = 'policy_lists_dataprefix.json'
 
 
+class PolicyListDataipv6prefix(PolicyList):
+    api_path = ApiPath('template/policy/list/dataipv6prefix')
+    store_path = ('policy_lists', 'DataIPv6Prefix')
+
+
+@register('policy_list', 'data-ipv6-prefix list', PolicyListDataipv6prefix)
+class PolicyListDataipv6prefixIndex(PolicyListIndex):
+    api_path = ApiPath('template/policy/list/dataipv6prefix', None, None, None)
+    store_file = 'policy_lists_dataipv6prefix.json'
+
+
 class PolicyListMirror(PolicyList):
     api_path = ApiPath('template/policy/list/mirror')
     store_path = ('policy_lists', 'Mirror')
@@ -1094,15 +1129,17 @@ class PolicyListTlocIndex(PolicyListIndex):
     store_file = 'policy_lists_tloc.json'
 
 
-class PolicyListDataipv6prefix(PolicyList):
-    api_path = ApiPath('template/policy/list/dataipv6prefix')
-    store_path = ('policy_lists', 'DataIPv6Prefix')
+# IP Prefix All (template/policy/list/ipprefixall) was purposely not included as it seems to collide with, meaning
+# error, IP Prefix (template/policy/list/prefix).
+class PolicyListPrefix(PolicyList):
+    api_path = ApiPath('template/policy/list/prefix')
+    store_path = ('policy_lists', 'Prefix')
 
 
-@register('policy_list', 'data-ipv6-prefix list', PolicyListDataipv6prefix)
-class PolicyListDataipv6prefixIndex(PolicyListIndex):
-    api_path = ApiPath('template/policy/list/dataipv6prefix', None, None, None)
-    store_file = 'policy_lists_dataipv6prefix.json'
+@register('policy_list', 'prefix list', PolicyListPrefix)
+class PolicyListPrefixIndex(PolicyListIndex):
+    api_path = ApiPath('template/policy/list/prefix', None, None, None)
+    store_file = 'policy_lists_prefix.json'
 
 
 class PolicyListIpv6prefix(PolicyList):
@@ -1116,6 +1153,17 @@ class PolicyListIpv6prefixIndex(PolicyListIndex):
     store_file = 'policy_lists_ipv6prefix.json'
 
 
+class PolicyListFQDN(PolicyList):
+    api_path = ApiPath('template/policy/list/fqdn')
+    store_path = ('policy_lists', 'FQDN')
+
+
+@register('policy_list', 'FQDN list', PolicyListFQDN, min_version='20.1')
+class PolicyListFQDNIndex(PolicyListIndex):
+    api_path = ApiPath('template/policy/list/fqdn', None, None, None)
+    store_file = 'policy_lists_fqdn.json'
+
+
 class PolicyListLocaldomain(PolicyList):
     api_path = ApiPath('template/policy/list/localdomain')
     store_path = ('policy_lists', 'LocalDomain')
@@ -1161,28 +1209,6 @@ class PolicyListTGApiKeyIndex(PolicyListIndex):
     store_file = 'policy_lists_tgapikey.json'
 
 
-class PolicyListFQDN(PolicyList):
-    api_path = ApiPath('template/policy/list/fqdn')
-    store_path = ('policy_lists', 'FQDN')
-
-
-@register('policy_list', 'FQDN list', PolicyListFQDN, min_version='20.1')
-class PolicyListFQDNIndex(PolicyListIndex):
-    api_path = ApiPath('template/policy/list/fqdn', None, None, None)
-    store_file = 'policy_lists_fqdn.json'
-
-# Shows up in SWAGGER for 20.1 but it is failing to be retrieved
-# class PolicyListDataPrefixFQDN(PolicyList):
-#     api_path = ApiPath('template/policy/list/dataprefixfqdn')
-#     store_path = ('policy_lists', 'DataPrefixFQDN')
-#
-#
-# @register('policy_list', 'data prefix FQDN list', PolicyListDataPrefixFQDN)
-# class PolicyListDataPrefixFQDNIndex(PolicyListIndex):
-#    api_path = ApiPath('template/policy/list/dataprefixfqdn', None, None, None)
-#    store_file = 'policy_lists_dataprefixfqdn.json'
-
-
 class PolicyListTransRules(PolicyList):
     api_path = ApiPath('template/policy/list/translationrules')
     store_path = ('policy_lists', 'TranslationRules')
@@ -1260,6 +1286,61 @@ class PolicyListTrunkGroupIndex(PolicyListIndex):
     store_file = 'policy_lists_trunkgroup.json'
 
 
+class PolicyAppProbe(PolicyList):
+    api_path = ApiPath('template/policy/list/appprobe')
+    store_path = ('policy_lists', 'AppProbe')
+
+
+@register('policy_list', 'app-probe class list', PolicyAppProbe, min_version='20.4')
+class PolicyAppProbeIndex(PolicyListIndex):
+    api_path = ApiPath('template/policy/list/appprobe', None, None, None)
+    store_file = 'policy_lists_appprobe.json'
+
+
+class PolicyListPort(PolicyList):
+    api_path = ApiPath('template/policy/list/port')
+    store_path = ('policy_lists', 'Port')
+
+
+@register('policy_list', 'port security list', PolicyListPort, min_version='20.4')
+class PolicyListPortIndex(PolicyListIndex):
+    api_path = ApiPath('template/policy/list/port', None, None, None)
+    store_file = 'policy_lists_port.json'
+
+
+class PolicyListProtocol(PolicyList):
+    api_path = ApiPath('template/policy/list/protocolname')
+    store_path = ('policy_lists', 'Protocol')
+
+
+@register('policy_list', 'protocol security list', PolicyListProtocol, min_version='20.4')
+class PolicyListProtocolIndex(PolicyListIndex):
+    api_path = ApiPath('template/policy/list/protocolname', None, None, None)
+    store_file = 'policy_lists_protocol.json'
+
+
+class PolicyListExpandedCommunity(PolicyList):
+    api_path = ApiPath('/template/policy/list/expandedcommunity')
+    store_path = ('policy_lists', 'ExpandedCommunity')
+
+
+@register('policy_list', 'expanded community list', PolicyListExpandedCommunity, min_version='20.5')
+class PolicyListExpandedCommunityIndex(PolicyListIndex):
+    api_path = ApiPath('/template/policy/list/expandedcommunity', None, None, None)
+    store_file = 'policy_lists_expanded_community.json'
+
+
+class PolicyListGeoLocation(PolicyList):
+    api_path = ApiPath('/template/policy/list/geolocation')
+    store_path = ('policy_lists', 'GeoLocation')
+
+
+@register('policy_list', 'geo location list', PolicyListGeoLocation, min_version='20.5')
+class PolicyListGeoLocationIndex(PolicyListIndex):
+    api_path = ApiPath('/template/policy/list/geolocation', None, None, None)
+    store_file = 'policy_lists_geo_location.json'
+
+
 #
 # Admin Settings
 #
@@ -1286,7 +1367,7 @@ class SettingsVbond(ConfigItem):
 #
 class EdgeCertificate(IndexConfigItem):
     api_path = ApiPath('certificate/vedge/list', 'certificate/save/vedge/list', None, None)
-    store_path = ('certificates', )
+    store_path = ('certificates',)
     store_file = 'edge_certificates.json'
     iter_fields = ('uuid', 'validity')
 
@@ -1332,28 +1413,28 @@ class EdgeCertificate(IndexConfigItem):
 #
 # Realtime items
 #
-@rt_register('system', 'status', 'System status')
+@op_register('system', 'status', 'System status')
 class SystemStatus(RealtimeItem):
     api_path = ApiPath('device/system/status', None, None, None)
     fields_std = ('state', 'cpu_user', 'cpu_system', 'mem_total', 'mem_free')
     fields_ext = ('disk_size', 'disk_used')
 
 
-@rt_register('bfd', 'sessions', 'BFD sessions')
+@op_register('bfd', 'sessions', 'BFD sessions')
 class BfdSessions(RealtimeItem):
     api_path = ApiPath('device/bfd/sessions', None, None, None)
     fields_std = ('system_ip', 'site_id', 'local_color', 'color', 'state')
     fields_ext = ('src_ip', 'src_port', 'dst_ip', 'dst_port')
 
 
-@rt_register('control', 'connections', 'Control connections')
+@op_register('control', 'connections', 'Control connections')
 class DeviceControlConnections(RealtimeItem):
     api_path = ApiPath('device/control/connections', None, None, None)
-    fields_std = ('system_ip',  'site_id', 'peer_type', 'local_color', 'remote_color', 'state')
+    fields_std = ('system_ip', 'site_id', 'peer_type', 'local_color', 'remote_color', 'state')
     fields_ext = ('private_ip', 'private_port', 'public_ip', 'public_port', 'instance', 'protocol', 'domain_id')
 
 
-@rt_register('control', 'local-properties', 'Control local-properties')
+@op_register('control', 'local-properties', 'Control local-properties')
 class DeviceControlLocalProperties(RealtimeItem):
     api_path = ApiPath('device/control/localproperties', None, None, None)
     fields_std = ('system_ip', 'site_id', 'device_type', 'organization_name', 'domain_id', 'port_hopped')
@@ -1361,7 +1442,34 @@ class DeviceControlLocalProperties(RealtimeItem):
                   'certificate_not_valid_after')
 
 
-@rt_register('interface', 'info', 'Interface info')
+@op_register('orchestrator', 'connections', 'Orchestrator connections')
+class OrchestratorConnections(RealtimeItem):
+    api_path = ApiPath('device/orchestrator/connections', None, None, None)
+    fields_std = ('system_ip', 'site_id', 'peer_type', 'local_color', 'state')
+    fields_ext = ('private_ip', 'private_port', 'public_ip', 'public_port', 'domain_id')
+
+
+@op_register('orchestrator', 'local-properties', 'Orchestrator local-properties')
+class OrchestratorLocalProperties(RealtimeItem):
+    api_path = ApiPath('device/orchestrator/localproperties', None, None, None)
+    fields_std = ('system_ip', 'device_type', 'organization_name', 'uuid', 'board_serial')
+    fields_ext = ('protocol', 'certificate_status', 'root_ca_chain_status', 'certificate_validity',
+                  'certificate_not_valid_after')
+
+
+@op_register('orchestrator', 'valid-vedges', 'Orchestrator valid WAN edges')
+class OrchestratorValidEdges(RealtimeItem):
+    api_path = ApiPath('device/orchestrator/validvedges', None, None, None)
+    fields_std = ('chassis_number', 'serial_number', 'validity')
+
+
+@op_register('orchestrator', 'valid-vsmarts', 'Orchestrator valid vSmarts')
+class OrchestratorValidVsmarts(RealtimeItem):
+    api_path = ApiPath('device/orchestrator/validvsmarts', None, None, None)
+    fields_std = ('serial_number', )
+
+
+@op_register('interface', 'info', 'Interface info')
 class InterfaceIpv4(RealtimeItem):
     api_path = ApiPath('device/interface', None, None, None)
     fields_std = ('vpn_id', 'ifname', 'af_type', 'ip_address', 'ipv6_address', 'if_admin_status', 'if_oper_status',
@@ -1369,22 +1477,22 @@ class InterfaceIpv4(RealtimeItem):
     fields_ext = ('tx_drops', 'rx_drops', 'tx_kbps', 'rx_kbps')
 
 
-@rt_register('app-route', 'stats', 'Application-aware route statistics')
+@op_register('app-route', 'stats', 'Application-aware route statistics')
 class AppRouteStats(RealtimeItem):
     api_path = ApiPath('device/app-route/statistics', None, None, None)
-    fields_std = ('index', 'remote_system_ip', 'local_color', 'remote_color', 'total_packets',
-                  'loss', 'average_latency', 'average_jitter')
+    fields_std = ('index', 'remote_system_ip', 'local_color', 'remote_color', 'total_packets', 'loss',
+                  'average_latency', 'average_jitter')
     fields_ext = ('mean_loss', 'mean_latency', 'mean_jitter', 'sla_class_index')
 
 
-@rt_register('app-route', 'sla-class', 'Application-aware SLA class')
+@op_register('app-route', 'sla-class', 'Application-aware SLA class')
 class AppRouteSlaClass(RealtimeItem):
     api_path = ApiPath('device/app-route/sla-class', None, None, None)
     fields_std = ('name', 'loss', 'latency', 'jitter')
-    fields_ext = ('index', )
+    fields_ext = ('index',)
 
 
-@rt_register('omp', 'summary', 'OMP summary')
+@op_register('omp', 'summary', 'OMP summary')
 class DeviceOmpSummary(RealtimeItem):
     api_path = ApiPath('device/omp/summary', None, None, None)
     fields_std = ('operstate', 'ompuptime', 'vsmart_peers', 'routes_received', 'routes_installed', 'routes_sent',
@@ -1392,29 +1500,170 @@ class DeviceOmpSummary(RealtimeItem):
     fields_ext = ('services_received', 'services_installed', 'services_sent', 'policy_received', 'policy_sent')
 
 
-@rt_register('omp', 'peers', 'OMP peers')
+@op_register('omp', 'peers', 'OMP peers')
 class DeviceOmpPeers(RealtimeItem):
     api_path = ApiPath('device/omp/peers', None, None, None)
     fields_std = ('peer', 'type', 'site_id', 'state')
     fields_ext = ('domain_id', 'up_time')
 
 
-@rt_register('tunnel', 'stats', 'Tunnel statistics')
+@op_register('omp', 'adv-routes', 'OMP advertised routes')
+class DeviceOmpRoutesAdv(RealtimeItem):
+    api_path = ApiPath('device/omp/routes/advertised', None, None, None)
+    fields_std = ('vpn_id', 'prefix', 'to_peer', 'color', 'ip', 'protocol', 'metric', 'preference')
+    fields_ext = ('tag', 'originator', 'site_id')
+
+
+@op_register('tunnel', 'stats', 'Tunnel statistics')
 class DeviceTunnelStats(RealtimeItem):
     api_path = ApiPath('device/tunnel/statistics', None, None, None)
     fields_std = ('system_ip', 'local_color', 'remote_color', 'tunnel_protocol', 'tunnel_mtu', 'tcp_mss_adjust')
     fields_ext = ('source_ip', 'dest_ip', 'source_port', 'dest_port')
 
 
-@rt_register('software', 'info', 'Software info')
+@op_register('software', 'info', 'Software info')
 class DeviceSoftware(RealtimeItem):
     api_path = ApiPath('device/software', None, None, None)
     fields_std = ('version', 'active', 'default')
-    fields_ext = ('confirmed', )
+    fields_ext = ('confirmed',)
 
 
-@rt_register('dpi', 'summary', 'DPI summary')
+@op_register('dpi', 'summary', 'DPI summary')
 class DeviceDpiSummary(RealtimeItem):
     api_path = ApiPath('device/dpi/summary', None, None, None)
     fields_std = ('status', 'current_flows', 'peak_flows', 'current_rate', 'peak_rate')
     fields_ext = ('flows_created', 'flows_expired')
+
+
+@op_register('arp', 'vedge', 'vEdge ARP table')
+class ArpVedge(RealtimeItem):
+    api_path = ApiPath('device/arp', None, None, None)
+    fields_std = ('vpn_id', 'if_name', 'ip', 'mac', 'state')
+
+    @classmethod
+    def is_in_scope(cls, device_model: str) -> bool:
+        return device_model not in CEDGE_SET
+
+
+@op_register('arp', 'cedge', 'cEdge ARP table')
+class ArpCedge(RealtimeItem):
+    api_path = ApiPath('device/arp', None, None, None)
+    fields_std = ('vpn_id', 'interface', 'address', 'hardware', 'mode')
+
+    @classmethod
+    def is_in_scope(cls, device_model: str) -> bool:
+        return device_model in CEDGE_SET
+
+
+@op_register('hardware', 'inventory', 'hardware inventory')
+class HardwareInventory(RealtimeItem):
+    api_path = ApiPath('device/hardware/inventory', None, None, None)
+    fields_std = ('hw_type', 'hw_description')
+    fields_ext = ('version', 'part_number', 'serial_number')
+
+
+@op_register('hardware', 'environment', 'hardware environment')
+class HardwareEnvironment(RealtimeItem):
+    api_path = ApiPath('device/hardware/environment', None, None, None)
+    fields_std = ('name', 'sensor_name', 'state')
+    fields_ext = ('current_reading', 'sensor_units')
+
+    @classmethod
+    def is_in_scope(cls, device_model: str) -> bool:
+        return device_model not in SOFT_EDGE_SET
+
+
+#
+# Bulk Statistics Items
+#
+@op_register('app-route', 'stats', 'Application-aware route statistics')
+class BulkAppRoute(BulkStatsItem):
+    api_path = ApiPath('data/device/statistics/approutestatsstatistics', None, None, None)
+    fields_std = ('local_system_ip', 'remote_system_ip', 'local_color', 'remote_color', 'total', 'loss', 'latency',
+                  'jitter', 'name')
+    fields_ext = ('tx_pkts', 'rx_pkts', 'tx_octets', 'rx_octets')
+    fields_to_avg = ('total', 'loss', 'latency', 'jitter')
+
+    @staticmethod
+    def time_series_key(sample: namedtuple) -> str:
+        return sample.name
+
+
+@op_register('interface', 'info', 'Interface info')
+class BulkInterfaceStats(BulkStatsItem):
+    api_path = ApiPath('data/device/statistics/interfacestatistics', None, None, None)
+    fields_std = ('vpn_id', 'interface', 'tx_kbps', 'rx_kbps', 'tx_pps', 'rx_pps')
+    fields_ext = ('rx_pkts', 'tx_pkts', 'rx_drops', 'tx_drops', 'rx_errors', 'tx_errors')
+    fields_to_avg = ('tx_kbps', 'rx_kbps', 'tx_pps', 'rx_pps')
+
+    @staticmethod
+    def time_series_key(sample: namedtuple) -> str:
+        return f"{sample.vdevice_name}_{sample.vpn_id}_{sample.interface}"
+
+
+@op_register('system', 'status', 'System status')
+class BulkSystemStats(BulkStatsItem):
+    api_path = ApiPath('data/device/statistics/devicesystemstatusstatistics', None, None, None)
+    fields_std = ('cpu_user', 'cpu_system', 'mem_util')
+    fields_ext = ('mem_used', 'mem_free', 'disk_used', 'disk_avail')
+    fields_to_avg = ('cpu_user', 'cpu_system', 'mem_util', 'mem_used', 'mem_free', 'disk_used', 'disk_avail')
+    field_conversion_fns = {
+        'cpu_system': abs,
+        'mem_util': lambda x: 100 * x
+    }
+
+
+#
+# Bulk State Items
+#
+@op_register('system', 'info', 'System info')
+class BulkSystemStatus(BulkStateItem):
+    api_path = ApiPath('data/device/state/SystemStatus', None, None, None)
+    fields_std = ('state', 'total_cpu_count', 'fp_cpu_count', 'linux_cpu_count', 'tcpd_cpu_count')
+    fields_ext = ('reboot_reason', 'reboot_type')
+
+
+@op_register('bfd', 'sessions', 'BFD sessions')
+class BulkBfdSessions(BulkStateItem):
+    api_path = ApiPath('data/device/state/BFDSessions', None, None, None)
+    fields_std = ('system_ip', 'site_id', 'local_color', 'color', 'state')
+    fields_ext = ('src_ip', 'src_port', 'dst_ip', 'dst_port', 'transitions', 'uptime_date')
+
+
+@op_register('control', 'connections', 'Control connections')
+class BulkControlConnections(BulkStateItem):
+    api_path = ApiPath('data/device/state/ControlConnection', None, None, None)
+    fields_std = ('system_ip', 'site_id', 'peer_type', 'local_color', 'remote_color', 'state')
+    fields_ext = ('private_ip', 'private_port', 'public_ip', 'public_port', 'instance', 'protocol', 'domain_id',
+                  'uptime_date')
+
+
+@op_register('control', 'local-properties', 'Control local-properties')
+class BulkControlLocalProperties(BulkStateItem):
+    api_path = ApiPath('data/device/state/ControlLocalProperty', None, None, None)
+    fields_std = ('system_ip', 'site_id', 'device_type', 'organization_name', 'domain_id', 'port_hopped')
+    fields_ext = ('protocol', 'tls_port', 'certificate_status', 'root_ca_chain_status', 'certificate_validity',
+                  'certificate_not_valid_after')
+
+
+@op_register('interface', 'vedge', 'vEdge interfaces')
+class BulkInterfaceVedge(BulkStateItem):
+    api_path = ApiPath('data/device/state/Interface', None, None, None)
+    fields_std = ('vpn_id', 'ifname', 'af_type', 'ip_address', 'ipv6_address', 'if_admin_status',
+                  'if_oper_status', 'desc')
+    fields_ext = ('mtu', 'hwaddr', 'speed_mbps', 'port_type')
+
+
+@op_register('interface', 'cedge', 'cEdge interfaces')
+class BulkInterfaceCedge(BulkStateItem):
+    api_path = ApiPath('data/device/state/CEdgeInterface', None, None, None)
+    fields_std = ('vpn_id', 'ifname', 'ip_address', 'ipv4_subnet_mask', 'ipv6_addrs', 'if_admin_status',
+                  'if_oper_status', 'description')
+    fields_ext = ('mtu', 'hwaddr', 'speed_mbps')
+
+
+@op_register('omp', 'peers', 'OMP peers')
+class BulkOmpPeers(BulkStateItem):
+    api_path = ApiPath('data/device/state/OMPPeer', None, None, None)
+    fields_std = ('peer', 'type', 'site_id', 'state')
+    fields_ext = ('domain_id',)
